;                    CREEM version 1.1 for the 8051
;        "Concurrent Realtime Embedded Executive for Microcontrollers"
;                   (c) copyright Barry Kauler 1997
;                         Grant Gorddard 1998
;
; Designed to implement projects designed by GOOFEE flow diagrams:
; see book "Flow Design for Embedded Systems", by Barry Kauler,
; 1997, R&D Books/Miller Freeman, http://www.rdbooks.com/
; See also GOOFEE Systems, http://www.goofee.com/ has a html file with an
; introduction to this version of CREEM.
;
; This version of CREEM support a Siemens SAE81C90 Controller Area Network
; device via a new macro REMOTEMSG for sending messages to remote boards.
; It has a queue ISR facility for receiveing messages and a dequeueing
; subroutine for extracting remote messages from the queue and adding them
; to the local buffer.
;
; Compiling requires Metalink assembler or Franklin A51 or similar.
; NOTE: one or more "#" at end of line indicate further notes at end of file
;
$MOD51       ; Open MOD51

;***************************************************************************
;   APP. PROGRAMMER TO SETUP TIMESLICE COUNTS, PRONUM AND CAN EXT.ADDRESS
;***************************************************************************
; EQUATES

TIMESLICE EQU  4     ;# entries to switcher before a context switch.
                     ;(4 gives about 0.5mSec for a 12MHz clock).
                     ;keep in the range 0--254 (not 255)

;>>> fMAXMSG EQU msgcnt.3               ;i.e. cnt=8 when msg buffer full.
;>>> Franklin A51 won't allow this, so use msgcnt.3 directly in code,
;>>> for 32-byte buffer. for 64-byte buff., use msgcnt.4 and change
;>>> range-checking in various places in scheduler.

CANHBYTE  EQU  0E0h  ;the upper 8 bits of the external CAN address (E0XXh)

ID1       EQU  01h   ;this board's processor number to accept remote msgs
ID2       EQU  04h   ;this board's extension number and RTR/DLC bits

; The identifier has been divided into two parts processor number (0--255)
; and the extension number (0--7).
; Together each board needs a unique combination of processor number
; that is combined with the extension number.  Together they make up the
; 11-bit unique identifier of the standard CAN specification V2.0 part A
; (see references).  RTR=0 and DLC=4 always, so ensure so in ID2 above.
;
;       -------------------------------------------------
; bits: |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
;       -------------------------------------------------
;  ID1: |               Processor number                |
;       -------------------------------------------------
;  ID2: |   Ext.Number    | RTR |         DLC           |
;       -------------------------------------------------

;***************************************************************************
; MACROS

MESSAGE MACRO fasync,inputs,term,fzone,node,datahi,datalo                ;##
    jb   msgcnt.3,$              ;loop until buffer not full.            ;##
    mov  a,#255                  ;prevent context switch.
    xch  a,22h                   ;/ tickcnt -- *tickcnt MUST BE AT ADDR 22h*
    mov  r0,23h                  ;ptail -- *ptail MUST BE AT ADDR 23h*
  IF fasync EQ 0    
    mov  @r0,#(128 + (fasync*64) + ((inputs-1)*8) + (term-1)) ;hdr1.
  ELSE
    mov  @r0,#((fasync*64) + ((inputs-1)*8) + (term-1)) ;hdr1.7 left clr.
  ENDIF
    dec  r0
    mov  @r0,#((fzone*128) + node)      ;hdr2.
    dec  r0
    mov  @r0,datahi
    dec  r0
    mov  @r0,datalo
    dec  r0
    mov  23h,r0                  ;update ptail. keep in limits --  *ptail*
    anl  23h,#00011111b          ;/  (address range 0110 0000 to   *ptail*
    orl  23h,#01100000b          ;/      0111 1111)                *ptail*
    inc  msgcnt                  ;how many msgs in buffer.
    xch  a,22h                   ;restore tickcnt.               *tickcnt*

; NOTE: Franklin A51 allows the labels tickcnt and ptail to be used in
; the macro.  However some other assembler macro preprocessors may not
; understand variables defined outside the macro.  The above should assemble
; everywhere.
; NOTE: fasync and fzone are flags -> 0 or 1. inputs and term -> 1--8.
;       node# -> 0--127. datahi and datalo -> 0--255.
MESSAGE ENDM


REMOTEMSG MACRO pronum,proext,fasync,inputs,term,fzone,node,datahi,datalo 
    mov  a,#255                  ;tickcnt = #255 to prevent context switch
    xch  a,22h                   ;tickcnt MUST be at 22h 
    push acc                     ;save tickcnt
    push dpl                     ;save DPTR
    push dph                     ;/
    mov  a,datalo                ;
    mov  dph,#CANHBYTE           ;load first 8 bits of external CAN address
    mov  dpl,#0FBh               ;byte3 of message object 15
    movx @dptr,a                 ;load datalo into byte3 of MO15
    mov  a,datahi                ;
    dec  dpl                     ;byte2 of MO15
    movx @dptr,a                 ;load datalo into byte2 of MO15
    mov  a,#((fzone*128) + node) ;formulate hdr2
    dec  dpl                     ;byte1 of MO15
    movx @dptr,a                 ;load hdr2 into byte1 of MO15
  IF fasync EQ 0                
      mov  a,#(128 + (fasync*64) + ((inputs-1)*8) + (term-1))  ;hdr1 (new=1)
  ELSE                          
      mov  a,#((fasync*64) + ((inputs-1)*8) + (term-1))        ;hdr1 (new=0)
  ENDIF
    dec  dpl                     ;byte0 of MO15
    movx @dptr,a                 ;load hdr1 into byte0 of MO15
    mov  a,#pronum               ;load processor number into acc
    mov  dpl,#5Eh                ;byte0 of MO15 descriptor bytes
    movx @dptr,a                 ;copy pronum into identifier bits.
    mov  a,#((proext*32) + 00000100b)  ;load processor extension, RTR & DLC
    inc  dpl                     ;byte1 of MO15 descriptor bytes
    movx @dptr,a                 ;copy last identifier bits, RTR=0 and DLC=4

; A quick check can be performed here if necessary
; Note that cjne will affect the carry flag!
;    mov  dpl,#10h                ;get CAN mode/status register
;    movx a,@dptr                 ;/
;    anl  a,#00011100b            ;ignore IM,RES,TC,RS,ADE bits
;    setb p1.0                    ;port 1 pin 0 will stay on if next loops
;    cjne a,#00000000b,$          ;dead-end loop if BS, RWL or TWL
;    clrb p1.0                    ;clear port pin if okay

    mov  dpl,#09h                ;load DPTR with transmit-request-set reg. 2
    mov  a,#10000000b            ;set bit TRS15
    movx @dptr,a                 ;send off MO15
    call checkTC                 ;wait until the message has been sent

    pop  dph                     ;restore DPTR
    pop  dpl                     ;/
    pop  acc                     ;restore acc containing tickcnt
    xch  a,22h                   ;restore tickcnt

;NOTE: pronum = remote processor number (0--255)
;      proext = remote processor extension (0--7)
;      other parameters as above for MESSAGE MACRO....
REMOTEMSG ENDM

;***************************************************************************
DSEG

ORG 20h              ;below this is the register banks.
flags:    DS 1       ;*must be bit-addressable*
msgcnt:   DS 1       ;how many msgs in buffer.         ;* / *
tickcnt:  DS 1       ;timer0 count. *MUST LEAVE THIS AT ADRESS 22h*
ptail:    DS 1       ;tail of msg buffer. *MUST LEAVE THIS AT ADDR 23h*
phead:    DS 1       ;head of msg buffer.
qtail:    DS 1       ;tail of remote message queue
qhead:    DS 1       ;head of remote message queue
swpsw:    DS 1       ;used exclusively by switcher.
swr0:     DS 1       ;/
watchdog: DS 1       ;=0 disabled. else #timeslices to timeout.

blocks:              ;context blocks.
first:    DS 8       ;/
second:   DS 8       ;/
third:    DS 8       ;/
fourth:   DS 8       ;/
blocksend:

pEMPTY EQU $-1       ;stack starts from end of internal variables.
                     ;grows from 4Ah (20h+2Ah) to 58h below = 14 bytes

ORG 58h              ;88d.
tempmsgs:  DS 1      ;used exclusively by scheduler.
tempmsgs1: DS 1      ;/   (if the stack ever runs into here,
tempmsgs2: DS 1      ;/    there is a "good" probability of
tempmsgs3: DS 1      ;/    no clash)
tempmsgs4: DS 1      ;/
tempmsgs5: DS 1      ;/
tempmsgs6: DS 1      ;/
tempmsgs7: DS 1      ;/

ORG 60h              ;96d.
; msg buffer is from here to address 127, 32 bytes, allowing 8 msgs.

;***************************************************************************
CSEG

; Note that the following code segments may be shifted to higher code space
; addresses except for the timer0 interrupt address which MUST be at 000Bh.

start1:

;    ORG 03h         ;int0 interrupt address triggered by CAN device
;    ajmp enqueue    ;
;    ljmp enqueue

;    ORG 0Bh         ;timer0 interrupt address. entered every 128*12/clk
;    ajmp switcher   ;                                  (clk in Hz ) sec.
;    ljmp switcher

    ORG 9000h       ;execution starting address
    ajmp start2

;    ORG 0032h      ;program code starting address
    ORG 90A1h

;***************************************************************************
;              APPLICATION PROGRAMMER TO SETUP THIS NODETABLE
;***************************************************************************
nodetable:
    ljmp node0      ;each entry takes up 3 bytes.
    ljmp node1      ;symbolic names can be whatever we want.
    ljmp node2      ;but must be in order of contiguous increasing node #.
    ljmp node3
    ljmp node4
    ljmp node5
    ljmp node6
    ljmp node7
    ljmp node8
    ljmp node9
    ljmp node10

;**************************************************************************
; INITIALISATION    (The program begins here...)

start2:
; Set up the 8051
      mov  sp,#pEMPTY       ;initialise stack.
      mov  r0,#07Fh         ;zeroise memory.
xst1: mov  @r0,#0           ;/
      djnz r0,xst1          ;/

; Initialising the CAN device
      mov  dph,#CANHBYTE    ;load first 8 bits of external CAN address
      mov  dpl,#10h         ;10h mode/status register
      mov  a,#03h           ;initmode=1, reset(hard)=1
      movx @dptr,a          ;put CAN into initialisation mode

      mov  dpl,#00h         ;CAN starting address XX00h
      mov  a,#00h           ;start counting from zero
      mov  b,#00h           ;duplicate copy of count acc
next: add  a,#12            ;skip 12 program bytes to the table
      movc a,@a+pc          ;load acc with value off following table
      movx @dptr,a          ;load value into CAN device
      inc  dpl              ;next address in the CAN device
      inc  b                ;increment copy of count
      mov  a,b              ;restore count to acc
      cjne a,#19,next       ;if counted 19 bytes go to next instruction
      sjmp clearbytes       ;jump over table to clearbytes

; Table of bytes to be loaded into the device from address XX00h to XX12h.
; Following configuration is for 50kb/s baudrate.
; (BL1=7Ah,BL2=41h,BRP=07h,TX1=000,TX0=110,OCM=00,IMSK=01,RIM1=FF,RIM2=7F)

;50KB   db 7Ah,41h,0F8h,07h,00,00,0FFh,7Fh,00,00,01,00,00,00,00,00,00,00,00    
;888kb
   db 58h,41h,0F8h,00h,00,00,0FFh,7Fh,00,00,01,00,00,00,00,00,00,00,00    


clearbytes:
; Clear bytes XX30h to XXFFh in the CAN device
      mov  dpl,#30h          ;start address
loop: mov  a,#00h            ;empty byte
      movx @dptr,a           ;write this to the CAN device
      inc  dpl               ;next address
      mov  a,dpl             ;copy low byte of the DPTR to A
      cjne a,#00h,loop       ;jump back to loop if next address is not 00h
      clr  psw.7             ;clear the carry flag

; Initialise the remote message queue
      mov  qhead,#50h        ;head of the queue is at MO 0 descriptor byte 0
      mov  qtail,#50h        ;tail of queue also at MO 0 descriptor byte 0
      mov  dpl,qhead         ;load dptr with head of queue
      mov  a,#ID1            ;load ID1 into MO 0, byte 0
      movx @dptr,a           ;/
      inc  dptr              ;dptr now points to MO 0 descriptor byte 1
      mov  a,#ID2            ;load ID2 into MO 0, byte 1
      movx @dptr,a           ;/

      mov  ptail,#127       ;initialise local message message buffer.
      mov  phead,#127       ;/

; Select which nodes are to run initially.  Presently in the first
; register bank, so for this thread first block *must* be activated.

      mov  r0,#first        ;same as #blocks.  *assigning node 1*
      mov  @r0,#1           ;load 1 into node field of block -- activates it.

; NOTE: Could start more blocks initially, but decided to restrict to
; node 1 only.  See the direct jmp to node 1 below...

      setb ie.0             ;allow external INT0 interrupt.
      setb ip.0             ;allow external interrupt priority over t0 int.
      setb ie.7             ;allow global interrupts.

; Now that the interrupts are initialised, wait 2 seconds for the other
; boards on the network to get initialied.  It is alright for other boards
; to send messages to this board at this stage because enqueue is ready.

      mov  tmod,#01         ;mode 1
      mov  b,#028h          ;40 (28h) * 50ms = 2000ms = 2 seconds
tim1: mov  th0,#04Bh        ;setup timer 0 with 50ms (4BFDh)
      mov  tl0,#0FDh        ; 
      clr  tf0              ;clear the rollover flag
      setb tr0              ;start timer 0
tim2: jnb  tf0,tim2         ;wait till rollover bit set
      clr  tr0              ;turn off timer 0
      djnz b,tim1           ;loop 40 (28h) times

; Set up timer0 now for the switcher using mode 2

      clr  tf0              ;clear the rollover flag
      setb ie.1             ;allow t0 interrupts. 
      mov  tl0,#00          ;timer 0 starting value
      mov  th0,#128         ;reload value
      mov  tmod,#02         ;mode 2
      setb tcon.4           ;starts timer 0

; Run node 1.  The initial context is the first block, first reg.bank,
; node 1, so jump directly...

      mov  dptr,#nodetable
      mov  a,#3             ;node 1 always second entry in nodetable.
      jmp @a+dptr           ;jump to node 1.

;***************************************************************************
; SCHEDULER

;scheduler finished (this is part of scheduler below)

sc9:  mov  a,r2             ;r2 has addr of current block.
      mov  r0,a
      mov  @r0,#0           ;deactivate this block.
      call dequeue          ;do a dequeue
      clr  flags.0          ;allow scheduler to be called again.
      mov  tickcnt,#0       ;release timeslice on next interrupt.
      mov  TL0,#250         ;WE MUST INTERRUPT FAST <<<<<<<<<<*****
      ajmp $                ;loop until timeslice ends.   ##EXIT SCHEDULER##
                            ;(note scheduler has another exit point below)
			       
scheduler:

; When a node exits, it jumps to here (in same context as the node).
; There can only be one instance of scheduler running...

      mov  r7,#0            ;used to hold flags for arrived msgs.(term 1--8)
      jb  flags.0,$         ;flags.0 set if scheduler already running.
      setb flags.0          ;scheduler is now running

; Firstly, do a dequeue of any remote messages

      call dequeue

; Find address of current block ...

      mov  a,psw
      anl  a,#00011000b     ;bits 3/4 select register bank/context block.
      add  a,#blocks        ;addr of current context-block.
      mov  r2,a             ;save in r2 for later use.
      mov  r1,a             ;fix for node posted msg to itself\
      mov  @r1,#127         ;         \( see notes at the end of the file )

; Look for a new msg in msg buffer...

sc20: mov  c,msgcnt.3       ;set flags.6 if buffer full, else clr.         #
      mov  flags.6,c        ;/                                             #
      mov  r0,phead
     
sc0:  mov  a,r0             ;ptr to hdr1.
      anl  a,#00011111b     ;/  (address range 0110 0000 to
      orl  a,#01100000b     ;/   0111 1111) keep ptail in limits.
      mov  r0,a
      cjne a,ptail,sc1      ;jmp if not at end of buffer.

; Come here if reached end of buffer or buffer empty or buffer full...

      jnb  flags.6,sc9      ;ignore first loop buffer full.                #
      clr  flags.6          ;/      (could replace with unconditional jmp) #
sc1:  mov  a,@r0            ;get hdr1
      jb   acc.7,sc2        ;set if a new sync msg.
sc5:  dec  r0
xc5:  dec  r0
      dec  r0
      dec  r0
      ajmp sc0
     
; new sync msg found...however don't launch if in same zone as another
; active node. It is also possible for new node to already be running.

sc2:  dec  r0               ;pts to hdr2 (zone/node)
      mov  a,#blocks
x1:   mov  r1,a
      mov  a,@r1            ;zone/node field of block.
      mov  b,@r0            ;hdr2 (zone/node) of new msg.
      jnb  b.7,xx1          ;jmp if no exclusion zone.
      anl  a,#01111000b     ;extract exclusion zone.
      anl  b,#01111000b     ;/
xx1:  anl  a,#01111111b     ;compare entire node#.(must screen off bit7)
      xrl  a,b
      jz   xc5              ;zone/node clash, ignore found new msg.(leave it marked as new)
      mov  a,r1            
      add  a,#8             ;addr of next block.
      cjne a,#blocksend,x1
      inc  r0               ;hdr1.

; ...fall down here if no zone or node clash was found.
; Now find out if reqd # msgs to launch node.

sc8:  mov  a,@r0            ;get hdr1.
      clr  acc.7            ;mark msg as no longer new.
      mov  @r0,a            ;/
      anl  a,#00111000b     ;extract #inputs reqd to fire node.
      rr   a                ;move field to rhs.
      rr   a                ;/
      rr   a                ;/       
      inc  a                ;field stored 0--7, modify to value 1--8.
     mov  b,a               ;temp save.
     dec  r0                ;get hdr2.
     mov  dph,@r0           ;/   (temp use of dph)
     inc  r0                ;back to hdr1.
     mov  r4,dph

; ...dph and r4 both have zone/node#, b has #inputs.
; Now search thru entire buffer...
; count how many sync msgs addressed to same node.
; Entered with b has reqd #, dph(& r4) has zone/node. dpl is tally...

      mov  c,msgcnt.3       ;special case. buff. full, phead==ptail,       #
      mov  flags.6,c        ;/   same as empty.                            #
      mov  r1,phead         ;hdr1.
      mov  dpl,#0           ;initialise count of msgs to same node.
sc3:  mov  a,r1
      anl  a,#00011111b     ;/  (address range 0110 0000 to
      orl  a,#01100000b     ;/   0111 1111) keep within limits.
      mov  r1,a
      cjne a,ptail,sc4
      jnb  flags.6,sc7      ;special case. jmp if finished search thru buff#
      clr  flags.6          ;/       (could replace with unconditional jmp)#
sc4:  mov  a,@r0            ;get hdr1.
      dec  r1               ;hdr2.
      jb   acc.6,x3         ;ignore async msgs.
      mov  a,@r1            ;get hdr2
      cjne a,dph,x3
      inc  dpl              ;same zone/node found.
      inc  r1               ;back to hdr1.                 *maybe redundant*
      mov  a,@r1            ;make sure marked not new.           * / *
      clr  acc.7            ;                                    * / *       
      mov  @r1,a            ;                                    * / *
      dec  r1               ;back to hdr2.                       * / *
x3:   dec  r1
      dec  r1
      dec  r1               ;hdr1 of next msg.
      ajmp sc3
sc7:  mov  a,dpl            ;must fire node if >= # reqd inputs.
      clr  c                ;/
      subb a,b              ;/
      jc   sc0              ;/ jmp to look for any more new msgs.

; Fall down here if found correct # msgs to fire a node.
; Before we extract msgs, need to allocate a new context.

;            *   *   *   *   *   *   *   *   *   *   *

; Find a free block...

      mov  a,#blocks
xt0:  mov  r1,a
      cjne @r1,#0,xt1       ;1st field is zone/node. =0 if empty.
      ajmp xt2              ;empty block found.
xt1:  add  a,#8
      cjne a,#blocksend,xt0 ;if fall below, no blocks available.

; So, use current block... already have in r2

      mov  a,r2
      mov  r1,a
xt2:  mov  r3,a
      clr  c
      subb a,#blocks        ;now pointing to reg.bank of new context.
      mov  r6,a

; Drop out of here with r1, r3 both pointing to new block,
; r2 pointing to current block, r6 ptg to reg.bank of new context.

;            *   *   *   *   *   *   *   *   *   *   *
      
; Found! launch node...
; Now have to search again and extract msgs to the node...
; Enter with dph(& r4) = zone/node that we want to find, r3 points to
; new block...

      mov  c,msgcnt.3       ;special case, buff. full.                     #
      mov  flags.6,c        ;/                                             #
      mov  r0,phead
sc11: mov  a,r0
      anl  a,#00011111b     ;/  (address range 0110 0000 to
      orl  a,#01100000b     ;/   0111 1111) keep within limits.
      mov  r0,a
      dec  r0               ;hdr2.
      cjne a,ptail,sc10
      jnb  flags.6,sc15     ;special case. jmp if end of buffer.           #
      clr  flags.6          ;/       (could replace with unconditional jmp)#
sc10: mov  a,@r0            ;hdr2
      xrl  a,dph            ;dph has original hdr2.
      jz   sc12             ;msg found.
sc13: dec  r0               ;datahi.
      dec  r0               ;datalo.
      dec  r0               ;next hdr1.
      ajmp sc11

; Extract found msg...

sc12: inc  r0               ;hdr1.

; Set flag in r7...
; note that finally we will put msgflags into b, in new context.

      mov  a,@r0            ;hdr1.
      anl  a,#00000111b     ;extract term# (1--8 stored as 0--7).
      inc  a
      mov  b,a
      mov  a,#1
te1:  djnz b,te2
      orl  a,r7             ;set flag in r7 (r7 in current context).
      mov  r7,a             ;/
      ajmp te3
te2:  rl   a
      ajmp te1
  
; Deliver data-msg to new context...
; Awkward problem if new block == current block, as would overwrite regs..

te3:  mov  a,r2             ;r2 pts to current block.
      mov  b,r3             ;r3 pts to new block.
      cjne a,b,xt3
      mov  r1,#tempmsgs
      ajmp xt4              ;leaves our current r0--7 alone for now.

; i.e., msg to term #1 goes to r0/1, #2 goes to r2/3, etc.

xt3:  mov  a,r6             ;pts to reg.bank of new context.
      mov  r1,a             ;/
xt4:  mov  a,@r0            ;hdr1.
      anl  a,#00000111b     ;extract term# (1--8 stored as 0--7).
      rl   a                ;as each data-msg is 2 bytes.
      add  a,r1             ;get addr to store data-msg.
      mov  r1,a             ;addr. of register to store data.
      dec  r0
      dec  r0               ;datahi.
      dec  r0               ;datalo. (Intel format, save at lower addr)
      mov  a,@r0
      mov  @r1,a            ;store lo-half of data.
      inc  r0               ;datahi.
      inc  r1               ;next register.
      mov  a,@r0
      mov  @r1,a            ;store hi-half of data.
                            
; We have just extracted a msg, now close the gap...
; r0 is pointing to datahi.
; Prevent a context switch here, as don't want a new msg appended while
; contracting the tail...

      inc  tickcnt          ;delay context switch by one tick.
      push tickcnt          ;save tickcnt
      mov  tickcnt,#255     ;add more ticks in case an INT0 occurs
      dec  r0               ;datalo
      mov  a,ptail          ;we will move msg from tail into gap.
      inc  a                ;point to datalo of tail msg.
      anl  a,#00011111b     ;/  (address range 0110 0000 to
      orl  a,#01100000b     ;/   0111 1111) keep within limits.
      mov  r1,a
      mov  a,@r1            ;move datalo.
      mov  @r0,a            ;/
      inc  r1
      inc  r0
      mov  a,@r1            ;move datahi.
      mov  @r0,a            ;/
      inc  r1
      inc  r0
      mov  a,@r1            ;move hdr2.
      mov  @r0,a            ;/
      inc  r1
      inc  r0
      mov  a,@r1            ;move hdr1.
      mov  @r0,a            ;/
      mov  ptail,r1         ;update contracted tail of buffer.
      dec  msgcnt           ;keep count of how many msgs in buffer.
      pop  tickcnt          ;restore tickcnt
      ajmp sc11             ;loop back and extract another msg.

;            *   *   *   *   *   *   *   *   *   *   *

; Now to activate the node.
; Enter with the node# in dph(& r4), r3 ptg to new block, r2 ptg to
; current block.
; Find correct entry in nodetable...

sc15: mov  a,r4
      jnb  acc.7,zo1        ;fzone set if zone active.
      clr  acc.7            ;screen off fzone, leaving node#.
zo1:  mov  b,#3             ;as jumptable entries 3 bytes.
      mul  ab               ;/ (a*b --> b:a)
      inc  a                ;rel.ptr to high-byte of jump address.
      mov  b,a
      mov  dptr,#nodetable
      movc a,@a+dptr        ;get hi-byte of jmp addr.
      xch  a,b
      inc  a                ;get lo-byte of jump address.
      movc a,@a+dptr        ;/
      mov  dpl,a
      mov  dph,b

; Load registers into new block...

      mov  a,r3             ;new block.
      mov  b,r2             ;current block.
      cjne a,b,xc1          ;r3==r2 means new block is current one.
      mov  r1,a
      mov  a,r4             ;r4 has new zone/node.
      mov  @r1,a            ;load new zone/node (block already activated)

; data-msgs not yet delivered. saved in tempmsgs...

      mov  b,r7             ;msgflags.
      mov  r0,tempmsgs
      mov  r1,tempmsgs1
      mov  r2,tempmsgs2
      mov  r3,tempmsgs3
      mov  r4,tempmsgs4
      mov  r5,tempmsgs5
      mov  r6,tempmsgs6
      mov  r7,tempmsgs7
      mov  a,#0
      call dequeue          ;call a dequeue now there is some room
      clr  flags.0          ;allows scheduler to be called again.
      mov  tickcnt,#0       ;release timeslice on next interrupt. 
      mov  TL0,#250         ;WE MUST INTERRUPT FAST <<<<<<<<<<*****
      jmp  @a+dptr          ;jump directly to new node.   ###EXIT###

xc1:  mov  a,r3             ;new block.
      mov  r1,a
      inc  r1               ;points to pc-lo field of new block.
      mov  @r1,dpl          ;program counter.
      inc  r1               ;points to pc-hi field of new block.
      mov  @r1,dph          ;program counter.
      inc  r1
      inc  r1
      inc  r1
      inc  r1               ;r1 now pts to b field of new block.
      mov  a,r7             ;r7 has msgflags.
      mov  @r1,a            ;/ b of new context will have msgflags.
      inc  r1               ;pts to psw field of new block.
      mov  a,r6             ;ptr to new reg.bank.
      mov  @r1,a            ;load new psw. (ptr matches with psw.3/4)
      mov  a,r3             ;points to zone/node field of new block.
      mov  r1,a
      mov  a,r4             ;zone/node.
      mov  @r1,a            ;activates new node!!!

; We have launched a new node, but need to check for more...
      ajmp sc20

;***************************************************************************
; SWITCHER

switcher:
      xch  a,tickcnt        ;save acc
      jnz  sw1
      mov  swpsw,psw        ;save psw.
      mov  swr0,r0          ;save r0 of current context.

; Watchdog...               ;default is this inactive.

      mov  a,watchdog
      jz  xw1               ;=0 means watchdog disabled.
      djnz watchdog,xw1
      setb flags.1          ;flag watchdog timeout.
xf1:  pop  acc              ;dump isr return addr.
      pop  acc              ;/
      mov  dptr,#nodetable  ;corrupts this context's dptr!
      mov  a,#2             ;get operand of ljmp instr.
      movc a,@a+dptr        ;pc-lo of node 0.
      push acc
      mov  a,#1
      movc a,@a+dptr        ;get pc-hi of node 0.
      push acc              ;addr of node 0 now on stack.
      ajmp sw3              ;go to node 0.

; Save current context...

xw1:  mov  a,psw
      anl  a,#00011000b     ;bits 3/4 select register bank/context block.
      add  a,#blocks        ;addr of current context-block.
      mov  r0,a

; 1st byte of block always left alone. zeroised after a node exits.

      inc  r0               ;note problem with hi-lo order of pc.
      inc  r0
      pop  acc              ;hi-half pc (we are popping ret. addr)
      mov  @r0,a            ;/ save.
      dec  r0
      pop  acc              ;lo-half pc ( / )
      mov  @r0,a            ;/ save.
      inc  r0
      inc  r0
      mov  @r0,dpl          ;save dptr
      inc  r0               ;/
      mov  @r0,dph          ;/
      inc  r0
      mov  @r0,tickcnt      ;save acc.
      inc  r0
      mov  @r0,b            ;save b
      inc  r0
      mov  @r0,swpsw        ;save psw.

;find next active block... (first field non-0)

sw2:  inc  r0               ;r0 now points to next block.
      clr  psw.5            ;flag to avoid endless looping if no block empty
xw3:  cjne r0,#blocksend,xw2
      mov  r0,#blocks
      jb   psw.5,xw4        ;allow to come here twice,
      setb psw.5            ;/                 then go to the error handler.
xw2:  cjne @r0,#0,xw5       ;jump if found active block. r0 pointing to it.
      mov  a,r0             ;point to next block.
      add  a,#8             ;/
      mov  r0,a             ;/
      ajmp xw3
xw4:  setb flags.2          ;go to error handler, node 0.
      ajmp xf1              ;/

; Note that the above will loop endlessly if no blocks are active.         
; Load new context...
; 1st byte of block always left alone. zeroised after a node exits.

xw5:  inc  r0
      mov  a,@r0            ;load pc-low (program counter) from block.
      push acc              ;/  (we are giving this interrupt routine
      inc  r0               ;        a new return address).
      mov  a,@r0            ;load pc-hi  ( / ) from block.
      push acc              ;/
      inc  r0
      mov  dpl,@r0          ;load dptr
      inc  r0               ;/
      mov  dph,@r0          ;/
      inc  r0
      mov  tickcnt,@r0      ;load acc.
      inc  r0
      mov  b,@r0            ;load b.
      inc  r0
      mov  a,@r0            ;load psw.

; Now, be careful... restore r0 of this context before changing psw...

      mov  r0,swr0          ;restore r0 of old context.
      mov  psw,a            ;now on new reg.bank for new context.
sw3:  mov  a,#TIMESLICE     ;reload timercnt.
sw1:  dec  a
      xch  a,tickcnt        ;restore acc and tickcnt.
      reti

;***************************************************************************
; ENQUEUE

; Upon receiving a RI interrupt, INT 0 vector address jumps here to enqueue
; the received message and inform the program that a message is waiting to
; be dequeued.  It has the highest priority interrupt therefore does not
; need to prevent a context switch.

enqueue:
      push acc              ;save acc
      push dpl              ;save DPTR
      push dph              ;/
      push psw              ;save program status word (preserve carry flag)
      mov  dph,#CANHBYTE    ;get CAN high byte for dph
      mov  dpl,qtail        ;load dptr with qtail
      mov  a,#00            ;empty byte
      movx @dptr,a          ;clear byte 0 of message object descriptor byte
      inc  dpl
      movx @dptr,a          ;clear byte 1 of message object descriptor byte
      inc  dpl              ;next message object descriptor byte 0
      mov  a,dpl
      cjne a,#5Eh,dontwrap  ;check if next message object is MO15 (5Eh/5Fh)
      mov  a,#40h           ;if so then message object is now MO 0 (40h)

dontwrap:
      mov  qtail,a          ;save new MO address to qtail
      cjne a,qhead,addid    ;check to see if queue is full (head=tail)
      setb flags.4          ;inform that buffer is now full.  Must dequeue!
; NOTE: This queue must be dequeued before the next remote message arrives
; otherwise the next message will not be accepted by this CAN device.
      ajmp clearreg

addid:
      mov  dpl,a            ;save new MO address to dptr
      mov  a,#ID1           ;load Identifier part 1 to MO descriptor byte 0
      movx @dptr,a          ;/
      inc  dpl
      mov  a,#ID2           ;load Identifier part 2 to MO descriptor byte 1
      movx @dptr,a          ;/

clearreg:
      mov  a,#00h           ;empty byte
      mov  dpl,#04h         ;load dptr with receive ready register 1
      movx @dptr,a          ;clear register
      inc  dpl              ;load dptr with receive ready register 2
      movx @dptr,a          ;clear register
      pop  psw              ;restore PSW 
      pop  dph              ;restore DPTR
      pop  dpl              ;/
      pop  acc              ;restore accumulator
      setb flags.7          ;inform that at least one message is waiting
      reti

;***************************************************************************
; DEQUEUE

; At the start and end of scheduler, this routine is executed to dequeue
; remote messages.

dequeue:
      jb flags.7,bufferfull ;continue only if there is a message to dequeue!
      ret                   ;return otherwise

bufferfull:
      jnb msgcnt.3,notfull  ;if the local message buffer is full then
      ret                   ;cannot dequeue, so return.

notfull:
      inc  tickcnt          ;prevent an immediate context switch
      push tickcnt          ;save tickcnt
      mov  tickcnt,#255     ;add more ticks in case an INT0 occurs
      push acc              ;save acc
      push dpl              ;save DPTR
      push dph              ;/
      push psw              ;save program status word (preserve carry flag)

; Find the address of MO data byte 0 from the address of descriptor byte 0.
; (data byte 0) = 4*(desc.byte) - 128  =  2*(2*(desc.byte) - 64) 

      mov  a,qhead          ;load MO descriptor byte 0 into acc
      rl   a                ;acc = 2*(desc.byte)
      anl  a,#00111111b     ;acc = 2*(desc.byte) - 64
      orl  a,#01000000b     ;/
      rl   a                ;acc = 2*(2*(desc.byte) - 64) = MO data byte 0
      mov  dpl,a            ;copy MO data byte 0 address into dptr
      mov  dph,#CANHBYTE    ;get CAN high byte for dph
      orl  dpl,#00000111b   ;add 7 to dpl, dptr is at address of data byte 7
      movx a,@dptr          ;read data byte 7 (load CAN shadow register)
      anl  dpl,#11111000b   ;subtract 7 from dpl, dptr is at data byte 0

      xch  a,R0             ;Contents of R0 moved into acc
      xch  a,ptail          ;Contents of R0 are now in ptail, ptail in acc
      mov  R0,a             ;a copy of ptail is now in R0

readnext:
      movx a,@dptr          ;read data byte from CAN into acc
      mov  @R0,a            ;copy byte onto the end of the message buffer
      dec  R0               ;move the tail pointer down one byte
      inc  dptr             ;increment dptr to next data byte
      mov  a,dpl
      anl  a,#00000111b     ;compare the last 3 bits to see if dptr has 
      cjne a,#04h,readnext  ;/ incremented 4 times ( read 4 bytes )
      xch  a,R0             ;new ptail value in R0 swapped with acc
      xch  a,ptail          ;new ptail value swapped for R0 contents
      mov  R0,a             ;R0 contents restored

; Check to see if remote buffer was full, if so then no identifier exists at
; ptail=phead, therefore must be added.

      mov  a,qhead          ;copy qhead pointer to acc
      jnb  flags.4,nextMO   ;check to see if remote buffer was full
      mov  dpl,a            ;save current qhead MO address to dptr
      mov  a,#ID1           ;load Identifier part 1 to MO descriptor byte 0
      movx @dptr,a          ;/
      inc  dptr             
      mov  a,#ID2           ;load Identifier part 2 to MO descriptor byte 1
      movx @dptr,a          ;/
      clr  flags.4          ;remote message buffer is no longer full

nextMO:
      inc  a                ;move qhead to the next MO descriptor byte 0
      inc  a                ;/
      cjne a,#5Eh,notMO15   ;check if next message object is MO15 (5Eh/5Fh)
      mov  a,#40h           ;if so then message object is now MO 0 (40h)

notMO15:
      mov  qhead,a          ;save new MO address to qhead
      cjne a,qtail,fixtail  ;check to see if queue empty (head=tail)
      clr  flags.7          ;if it is empty, clear message waiting flag

fixtail:
      anl  ptail,#00011111b ;keep ptail in the address range 0110 0000 to
      orl  ptail,#01100000b ;/ 0111 1111
      inc  msgcnt           ;update message count
      pop  psw              ;restore PSW 
      pop  dph              ;restore DPTR
      pop  dpl              ;/
      pop  acc              ;restore accumulator
      pop  tickcnt          ;restore tickcnt value
      dec  tickcnt          ;decrement tickcnt (see above at 'notfull:')
      jb flags.7,bufferfull ;continue only if there is a message to dequeue!
      ret                   ;return otherwise

;***************************************************************************
; CHECKTC

; This subroutine waits for the message to be transmitted correctly and then
; clears the identifier from message object 15 so it cannot receive messages

checkTC:
    mov  dpl,#10h                ;load DPTR with CAN mode/status register
    push psw
tcwait:
    mov  a,#0Ah                  ;loop 20 machine cycles
    djnz acc,$                   ;jump to itself 10 times (2 cycles each)
    movx a,@dptr                 ;read CAN mode/status register
    anl  a,#00100000b            ;ignore all bits except TC
    cjne a,#00100000b,tcwait     ;wait another 20 machine cycles if TC=0
    pop psw

; Now that transmission complete (TC) has been set, the message identifier
; should be removed to stop it receiving messages.

    mov  a,#00                   ;empty byte
    mov  dpl,#5Eh                ;byte0 of MO15 descriptor bytes
    movx @dptr,a                 ;clear descriptor byte 0
    inc  dpl
    movx @dptr,a                 ;clear descriptor byte 1
    mov  dpl,qtail               ;load dptr with qtail
    ret

;***************************************************************************
;                USER APPLICATION CODE PAST HERE
;***************************************************************************

;the only thing application programmer has to worry about above,
;is setting up the jumptable, & how many tick counts between timeslices.
;nodes may use acc, b, dptr, psw, as these are saved in a
;context switch. nodes may also use r0--r7 of current bank.
;on entry to a node, the msgs are in r0--r7, corresponding to
;terminal numbers 1--4.  b register bits 0--7 are correspondingly set
;if a msg is present (bit 0 for msg at term #1, etc).

;note that the message format allows 8 terminals into a node, but
;as data fields are 16-bit, the use of a register bank to pass them to
;a node only allows 4 max.
;-- that is, when designing the flow diagram, only use terminals 1--4
;-- a node with multiple rings can reuse the same terminal numbers on
;   each ring, so effectively can have a lot more wires into a node.
;
;.......................................................................
node0:                     ;error handler.
      jnb  flags.1,n0a     ;set if watchdog timeout.
n0a:    
      jnb  flags.2,n0b     ;set if no nodes running.
n0b:
      ajmp scheduler
;........................................................................
node1:                              ;starts initially.
      mov p1,#00h                   ;clear port1
      MESSAGE 0,1,1,0,2,#11h,#22h   ;msg to node 2, term 1.
      MESSAGE 0,1,1,0,6,#11h,#22h   ;msg to node 6, term 1.
      ajmp scheduler
;........................................................................
node2:
      cpl  p1.2                     ;toggle port1.2
      MESSAGE 0,1,1,0,3,#55h,#66h   ;msg to node 3, term 1.
      MESSAGE 0,1,1,0,4,#77h,#88h   ;msg to node 4, term 1.
      ajmp scheduler
;........................................................................
node3:
      cpl  p1.3                     ;toggle port1.3
      mov  b,#6                     ;outer delay loop.
n3b:  mov  a,#255                   ;inner delay loop.
n3a:  djnz acc,n3a                  ; /
      djnz b,n3b
      MESSAGE 0,2,1,0,5,#12h,#34h   ;msg to node 5, term 1.
      ajmp scheduler
;........................................................................
node4:
      cpl  p1.4                     ;toggle port1.4
      mov  b,#3                     ;outer delay loop
n4b:  mov  a,#255                   ;inner delay loop.
n4a:  djnz acc,n4a                  ; /
      djnz b,n4b
      MESSAGE 0,2,2,0,5,#56h,#78h   ;msg to node 5, term 2.
      ajmp scheduler
;........................................................................
node5:
      cpl  p1.5                           ;toggle port1.5
      REMOTEMSG 24h,0,0,1,1,0,2,#77h,#77h ;msg to proc 24, extension 0,
                                          ;/node 2, term 1.                                            
      ajmp scheduler
;........................................................................
node6:
      cpl  p1.6                    ;toggle port1.6
      mov  b,#255                  ;outer delay loop.
n6b:  mov  a,#255                  ;inner delay loop.
n6a:  djnz acc,n6a                 ;/
      djnz b,n6b                   ;/
      MESSAGE 0,1,1,0,6,#01h,#02h  ;msg to itself.
      ajmp scheduler
;........................................................................
node7:
      nop
      ajmp scheduler
;........................................................................
node8:
      nop
      ajmp scheduler
;........................................................................
node9:
      nop
      ajmp scheduler
;........................................................................
node10:
      nop
      ajmp scheduler 
	
;note node0 isn't executed. the convention is that it is the error
;handler, and only executes when an error msg posted to it. set fzone
;flag when post msg to node 0, as hdr2=0 is not allowed.
;CREEM enters node 0 directly when it detects an error, with flag set
;in flags.
;
;***************************************************************************
;       END OF USER APPLICATION.       CREEM DETAILS PAST HERE.
;***************************************************************************
;
;SWITCHER
;********
;uses psw.3/4 to control current context block. this also allocates
;a register bank for each context. so we only have 4 concurrent threads.

;context block stores, in order of increasing addresses:
; zone/node, pc-lo, pc-hi, dpl, dph, acc, b, psw.
;switcher does not touch the zone/node entry in context blocks,
;as this is the responsibility of the scheduler and START function.

;therefore, a node may use acc, b, dptr, psw, as well as r0--r7.

;there is only one stack!!! sp is not saved for each context, which
;imposes some limitations.  i.e., a node may not use the stack to
;save temporary data, nor can there be any function calls!!!
; -- the solution is to prevent a context switch while stack is 
;being manipulated. simple code to do this:
;  mov  a,#255
;  xch  a,tickcnt
;  ...                          ;use stack.
;  xch  a,tickcnt               ;if acc wasn't changed, will restore tickcnt.

;or, a context switch can be delayed by one timer0 interrupt (one clock
;tick), by:  inc tickcnt

;it is very easy for an application to release context: mov tickcnt,#0
;there is still a variable delay before next timer0 interrupt,
;but switcher will then perform an immediate context switch.

;.........................................................................
;SCHEDULER
;*********
;when a node exits, it jumps directly to the scheduler.
;sheduler can use the register bank of the node that has just exited.
;therefore, scheduler continues to get timeslices, but when exits,
;will free the context, i.e. reg.bank and context block.

;scheduler will find a new msg and check against all other msgs to see
;if node is ready to fire. if not, marks msg as old. if yes, extracts
;messages to r0--r7 and set flags in b.

;msgs are in the format hdr1, hdr2, datahi, datalo.
;msgs are stored in msg buffer with hdr1 at highest address.
;phead points to hdr1 of msg at head of buffer. ptail points one
;below datalo of msg at tail of buffer.
;it is a circulating buffer, header grows to higher addresses, and
;wraps. the buffer is constrained in the address range 60h--7Fh.
;SEE BELOW.

;if we find reqd # of sync msgs have arrived to fire a node, we
;do a complete search again through the buffer, extracting the msgs.
;when a msg is extracted, a space becomes empty in the buffer.
;the msg at tail is moved up to fill the gap, and tail contracts.

;if a node exits, having posted a msg to itself, it will re-fire, as in
;the example of node6. the logic at "x1:" prevents this -- the problem is
;that the scheduler is in the context of the node that just exited, so
;CREEM thinks node6 is still running. node6 will re-fire later, when
;another node exits, but this is not satisfactory.
; -- the solution is that the scheduler changes the zone/node field in
;the current block to 127 (7Fh). this is also useful for debugging, as
;a "watch" on the blocks first, second, third, and fourth shows when
;the scheduler is executing. however it means that node 127 cannot be used.

;.........................................................................
;WATCHDOG
;********
;very simple, yet powerful. watchdog is a variable, that defaults
;to value 0, which disables the watchdog.
;switcher ignores watchdog if it is zero. however, if it is non-0
;switcher will decrement it every timeslice, and if it decrements
;to zero, switcher will go to node 0, the error handler.

;if your application puts a non-0 value into watchdog, and keeps
;putting it in regularly, fine. this is just like any normal
;watchdog. however, if your application puts zero into watchdog,
;watchdog becomes inactive. so, you can switch it on and off.

;another thing that is beyond simple watchdogs is that when the
;watchdog times out, switcher does a reti to node 0, but still in
;the context when it timed out. so, node0 can examine the registers,
;and, if necessary, the other context blocks.
;by looking at the context blocks, node0 will know exactly what
;nodes are running.

;an extremely interesting aspect of this is that on the next
;timeslice assigned to the context node0 is in, switcher will
;go through as normal and will not reti to node0.
; -- i.e. the watchdog only times out once, then is disabled.
;thus, node0 could take some corrective action then just loop
;until it loses the timeslice, and everything will carry on as though
;nothing happened.

;how to start the watchdog:  mov watchdog,#200 ;200mSec timeout.
;must re-execute this instruction within 200mSec intervals. (12MHz)
;how to disable watchdog:    mov watchdog,#0   ;never times out.

;.........................................................................
;MESSAGE BUFFER
;**************
;the format is:
;    7    6    5    4    3    2    1    0  
;  |new|async|  #inputs    |    term#     |      HDR1
;  | 0 |            node#                 |      HDR2
;  |                                      |      DATAHI
;  |                                      |      DATALO

;however, HDR2 has an alternative format if bit7 is set:
;  | 1 |     excl. zone    |              |      HDR2

;meaning of fields:
; new          used by the O.S. marks a new sync msg delivered to buffer.
;fasync        bit6 hdr1 =1 if msg is asynchronous, =0 if synchronous.
;#inputs       a node can have up to 8 sync inputs (stored as 0--7).
;term#         terminal # on node.  1--8, stored as 0--7.
;fzone         bit7 of hdr2, if =1 no concrurrency allowed in same zone.
;zone          only one node in same zone allowed to exec. concurrently.
;node#         node that msg is going to. 0--127.

;if fzone=1, the node number space is partitioned into groups of 8 nodes
;belonging to a zone.  hdr2.3--.6 is part of the node number but designates
;the zone.  that is, nodes 0--7 belong in same exclusion zone, 8--15 in
;next exclusion zone, and so on, giving 16 exclusion zones.

;you may wonder why put #inputs in each msg -- isn't this a constant for
;a node, hence can be in a lookup table? no, with the rationale behind
;GOOFEE diagrams, this is not a constant -- an Iterative node has multiple
;rings, one of which is active during execution, and each ring has a
;different number of inputs.
;note also that #inputs refers to number of *synchronous* msgs required to
;fire the node. GOOFEE diagrams also have asynchronous msgs -- these are
;ignored for scheduling purposes, but still get delivered when the node
;fires.

;the msg is stored in the buffer with hdr1 at highest address.
;in this version of CREEM, the buffer is very small:
;             |  SFRs   |
;   (128) 80h |_________|
;         7Fh |         |  <<<top limit of buffer
;             |         |
;    phead -->|  hdr1   |
;             |  hdr2   |
;             |  datahi |
;             |  datalo |
;    ptail -->|         |
;             |         |
;             |         |
;         60h |_________|  <<<bottom limit of buffer.

;the above example only has one msg in the buffer. note that phead
;points to hdr1 of msg at head of buffer, while ptail points one-below
;tail of the buffer.
;this is a circulating buffer. local new msgs get appended to the tail,
;and when msgs are extracted, it is the tail that contracts to fill
;the gaps.
;in future if a network module is added, an interrupt routine can append
;a remote new input msg at the head of the buffer, without interfering
;with local msgs being delivered at the same time.
;that is, interrupts never need to be turned off with this OS.

;the buffer is 32 bytes, allowing 8 msgs. design the
;system carefully to keep within this.
;when the buffer is empty *or* full, ptail==phead, so I need some other
;mechanism to check when buffer empty or full. I have used 
;"fMAXMSG EQU msgcnt.3" which means msgcnt of 8 is buffer full(bit3 set).
;to specify what the buffer can hold, and "msgcnt" to specify how many
;msgs currently in buffer.

;the system degrades gracefully if too many msgs posted to the buffer.
;MESSAGE macro refuses to post into a full buffer, and loops until
;space is available.


;NOTES
;*****

;#
;if the entire msg buffer is used, i.e. can fill right up, there is the
;problem that ptail==phead if full *or* empty. This requires the extra
;checking, marked with "#".
;if a full buffer is one-less, i.e. 7 msgs, then the MESSAGE macro will
;only post a msg if msgcnt less than 7. the situation of ptail==phead
;when buffer full doesn't arise, so all the code marked with "#" isn't
;needed -- delete, and where there is a conditional test replace with an
;unconditional jmp.
;only problem is wastes 4 bytes of precious internal RAM.

;##
;for async msgs, the #inputs and fzone fields of a msg are redundant.
; -- maybe in the future they could hold extra info for async msgs.
; -- maybe also the new/async fields = 1/1 could designate a different
;type of message (sync. msgs = 1/0, async = 0/0 -- i.e. the new field
;isn't set for async msgs).


;.........................................................................
    END
